"""
Calculate and View Aerosol Percentages
--------------------------------------

Calculate the percentages of different aerosols in a Aerosol
Chemical Speciation (AOS) monitor dataset and view the percentages
in a pie chart.

Written: Zach Sherman

"""

from arm_test_data import DATASETS
import matplotlib.pyplot as plt

import act
from act.io.arm import read_arm_netcdf

# Read an ARM AOS dataset
filename = DATASETS.fetch('sgpaosacsmE13.b2.20230420.000109.nc')
ds = read_arm_netcdf(filename)

# Let us print out the fields in the dataset and see what it contains.
print(ds.data_vars.keys())

# Knowing what fields the dataset contains, let's create a list of fields
# to use in the plot.

fields = ['sulfate', 'ammonium', 'nitrate', 'chloride']

# We also want to provide some keyword arguments to avoid invalid data such
# as negative values.
threshold = 0.0
fill_value = 0.0

# Create a DistributionDisplay object to compare fields
display = act.plotting.DistributionDisplay(ds)

# We can set one of the slices to explode and give it a nice shadow.
explode = (0, 0.1, 0, 0)
shadow = True

# Create a pie chart using the fields list. The percentages of the
# fields will be calculated using act.utils.calculate_percentages.
display.plot_pie_chart(
    fields,
    threshold=threshold,
    fill_value=fill_value,
    explode=explode,
    shadow=True,
)
plt.show()



"""
Modules for reading in NOAA PSL data.
"""


import pandas as pd
import xarray as xr

from .text import read_csv


def read_neon_csv(files, variable_files=None, position_files=None):
    """
    Reads in the NEON formatted csv files from local paths or urls
    and returns an Xarray dataset.

    Parameters
    ----------
    filepath : list
        Files to read in
    variable_files : list
        Name of variable files to read with metadata. Optional but the Dataset will not
        have any metadata
    position_files : list
        Name of file to read with sensor positions.  Optional, but the Dataset will not
        have any location information

    Return
    ------
    ds : xarray.Dataset
        Standard Xarray dataset

    """

    # Raise error if empty list is passed in
    if len(files) == 0:
        raise ValueError('File list is empty')
    if isinstance(files, str):
        files = [files]

    # Read in optional files
    multi_ds = []
    if variable_files is not None:
        if isinstance(variable_files, str):
            variable_files = [variable_files]
        df = pd.read_csv(variable_files[0])

    if position_files is not None:
        if isinstance(position_files, str):
            position_files = [position_files]
        loc_df = pd.read_csv(position_files[0], dtype=str)

    # Run through each file and read into a dataset
    for i, f in enumerate(files):
        ds = read_csv(f)
        # Create standard time variable
        time = [pd.to_datetime(t).replace(tzinfo=None) for t in ds['startDateTime'].values]
        ds['time'] = xr.DataArray(data=time, dims=['index'])
        ds['time'].attrs['units'] = ''
        ds = ds.swap_dims({'index': 'time'})
        ds = ds.drop_vars('index')

        # Add some metadata
        site_code = f.split('/')[-1].split('.')[2]
        resolution = f.split('/')[-1].split('.')[9]
        hor_loc = f.split('/')[-1].split('.')[6]
        ver_loc = f.split('/')[-1].split('.')[7]
        ds.attrs['_sites'] = site_code
        ds.attrs['averaging_interval'] = resolution.split('_')[-1]
        ds.attrs['HOR.VER'] = hor_loc + '.' + ver_loc

        # Add in metadata from the variables file
        if variable_files is not None:
            for v in ds:
                dummy = df.loc[(df['table'] == resolution) & (df['fieldName'] == v)]
                ds[v].attrs['units'] = str(dummy['units'].values[0])
                ds[v].attrs['long_name'] = str(dummy['description'].values[0])
                ds[v].attrs['format'] = str(dummy['pubFormat'].values[0])

        # Add in sensor position data
        if position_files is not None:
            dloc = loc_df.loc[loc_df['HOR.VER'] == hor_loc + '.' + ver_loc]
            idx = dloc.index.values
            if len(idx) > 0:
                if len(loc_df['referenceLatitude'].values) > 1:
                    ds['lat'] = xr.DataArray(data=float(loc_df['referenceLatitude'].values[idx][0]))
                    ds['lon'] = xr.DataArray(
                        data=float(loc_df['referenceLongitude'].values[idx][0])
                    )
                    ds['alt'] = xr.DataArray(
                        data=float(loc_df['referenceElevation'].values[idx][0])
                    )
                else:
                    ds['lat'] = xr.DataArray(data=float(loc_df['referenceLatitude'].values[idx]))
                    ds['lon'] = xr.DataArray(data=float(loc_df['referenceLongitude'].values[idx]))
                    ds['alt'] = xr.DataArray(data=float(loc_df['referenceElevation'].values[idx]))
                variables = [
                    'xOffset',
                    'yOffset',
                    'zOffset',
                    'eastOffset',
                    'northOffset',
                    'pitch',
                    'roll',
                    'azimuth',
                    'xAzimuth',
                    'yAzimuth',
                ]
                for v in variables:
                    if len(loc_df[v].values) > 1:
                        ds[v] = xr.DataArray(data=float(loc_df[v].values[idx][0]))
                    else:
                        ds[v] = xr.DataArray(data=float(loc_df[v].values[idx]))
        multi_ds.append(ds)

    ds = xr.merge(multi_ds)

    return ds



"""
Plotting state variables
------------------------

Simple examples for plotting state variable using flag_values
and flag_meanings.

Author: Ken Kehoe

"""

from arm_test_data import DATASETS
from matplotlib import pyplot as plt

from act.io.arm import read_arm_netcdf
from act.plotting import TimeSeriesDisplay

# ---------------------------------------------------------------------- #
# This example will create a plot of the detection status time dimentioned
# varible and set the y axis to the string values defined in flag_meanings
# instead of plotting the flag values.
# ---------------------------------------------------------------------- #

# Read in data to plot. Only read in the variables that will be used.
variable = 'detection_status'
filename_ceil = DATASETS.fetch('sgpceilC1.b1.20190101.000000.nc')
ds = read_arm_netcdf(filename_ceil, keep_variables=[variable, 'lat', 'lon', 'alt'])

# Clean up the variable attributes to match the needed internal standard.
# Setting override_cf_flag allows the flag_meanings to be rewritten using
# the better formatted attribute values to make the plot more pretty.
ds.clean.clean_arm_state_variables(variable, override_cf_flag=True)

# Creat Plot Display by setting figure size and number of plots
display = TimeSeriesDisplay(ds, figsize=(12, 8), subplot_shape=(1,))

# Plot the variable and indicate the day/night background should be added
# to the plot.
# Since the string length for each value is long we can ask to wrap the
# text to make a better looking plot by setting the number of characters
# to keep per line with the value set to y_axis_flag_meanings. If the
# strings were short we can just use y_axis_flag_meanings=True.
display.plot(variable, day_night_background=True, y_axis_flag_meanings=18)

# Display plot in a new window
plt.show()

# ----------------------------------------------------------------------- #
# This example will plot the 2 dimentional state variable indicating
# the cloud type classificaiton. The plot will use the correct formatting
# for x and y axis, but will show a colorbar explaining color for each value.
# ----------------------------------------------------------------------- #
# Read in data to plot. Only read in the variables that will be used.
variable = 'cloud_phase_hsrl'
filename_cloud = DATASETS.fetch('nsacloudphaseC1.c1.20180601.000000.nc')
ds = read_arm_netcdf(filename_cloud)

# Clean up the variable attributes to match the needed internal standard.
ds.clean.clean_arm_state_variables(variable, override_cf_flag=True)

# Creat Plot Display by setting figure size and number of plots
display = TimeSeriesDisplay(ds, figsize=(12, 8), subplot_shape=(1,))

# We need to pass in a dictionary containing text and color information
# for each value in the data variable. We will need to define what
# color we want plotted for each value but use the flag_values and
# flag_meanings attribute to supply the other needed information.
y_axis_labels = {}
flag_colors = ['white', 'green', 'blue', 'red', 'cyan', 'orange', 'yellow', 'black', 'gray']
for value, meaning, color in zip(
    ds[variable].attrs['flag_values'], ds[variable].attrs['flag_meanings'], flag_colors
):
    y_axis_labels[value] = {'text': meaning, 'color': color}

# Create plot and indicate the colorbar should use the defined colors
# by passing in dictionary to colorbar_lables.
# Also, since the test to display on the colorbar is longer than normal
# we can adjust the placement of the colorbar by indicating the adjustment
# of horizontal locaiton with cbar_h_adjust.
display.plot(variable, colorbar_labels=y_axis_labels, cbar_h_adjust=0)

# To provide more room for colorbar and take up more of the defined
# figure, we can adjust the margins around the initial plot.
display.fig.subplots_adjust(left=0.08, right=0.88, bottom=0.1, top=0.94)

# Display plot in a new window
plt.show()



import numpy as np

import act


def test_doppler_lidar_winds():
    # Process a single file
    dl_ds = act.io.arm.read_arm_netcdf(act.tests.sample_files.EXAMPLE_DLPPI)
    result = act.retrievals.doppler_lidar.compute_winds_from_ppi(dl_ds, intensity_name='intensity')
    assert np.round(np.nansum(result['wind_speed'].values)).astype(int) == 1570
    assert np.round(np.nansum(result['wind_direction'].values)).astype(int) == 32635
    assert result['wind_speed'].attrs['units'] == 'm/s'
    assert result['wind_direction'].attrs['units'] == 'degree'
    assert result['height'].attrs['units'] == 'm'
    dl_ds.close()

    # Process multiple files
    dl_ds = act.io.arm.read_arm_netcdf(act.tests.sample_files.EXAMPLE_DLPPI_MULTI)
    del dl_ds['range'].attrs['units']
    result = act.retrievals.doppler_lidar.compute_winds_from_ppi(dl_ds)
    assert np.round(np.nansum(result['wind_speed'].values)).astype(int) == 2854
    assert np.round(np.nansum(result['wind_direction'].values)).astype(int) == 64986
    dl_ds.close()



"""
Sample data file for use in testing. These files should only
be used for testing ACT.

"""


from arm_test_data import DATASETS

# Single files
EXAMPLE_MPEG = DATASETS.fetch('nsacamskyradmovieC1.a1.20240401.100300.mpg')
EXAMPLE_MET1 = DATASETS.fetch('sgpmetE13.b1.20190101.000000.cdf')
EXAMPLE_MET_SAIL = DATASETS.fetch('gucmetM1.b1.20230301.000000.cdf')
EXAMPLE_MET_CSV = DATASETS.fetch('sgpmetE13.b1.20210401.000000.csv')
EXAMPLE_METE40 = DATASETS.fetch('sgpmetE40.b1.20190508.000000.cdf')
EXAMPLE_CEIL1 = DATASETS.fetch('sgpceilC1.b1.20190101.000000.nc')
EXAMPLE_SONDE1 = DATASETS.fetch('sgpsondewnpnC1.b1.20190101.053200.cdf')
EXAMPLE_LCL1 = DATASETS.fetch('met_lcl.nc')
EXAMPLE_ANL_CSV = DATASETS.fetch('anltwr_mar19met.data')
EXAMPLE_VISST = DATASETS.fetch('twpvisstgridirtemp.c1.20050705.002500.nc')
EXAMPLE_MPL_1SAMPLE = DATASETS.fetch('sgpmplpolfsC1.b1.20190502.000000.cdf')
EXAMPLE_IRT25m20s = DATASETS.fetch('sgpirt25m20sC1.a0.20190601.000000.cdf')
EXAMPLE_NAV = DATASETS.fetch('marnavM1.a1.20180201.000000.nc')
EXAMPLE_AOSMET = DATASETS.fetch('maraosmetM1.a1.20180201.000000.nc')
EXAMPLE_DLPPI = DATASETS.fetch('sgpdlppiC1.b1.20191015.120023.cdf')
EXAMPLE_BRS = DATASETS.fetch('sgpbrsC1.b1.20190705.000000.cdf')
EXAMPLE_AERI = DATASETS.fetch('sgpaerich1C1.b1.20190501.000342.nc')
EXAMPLE_IRTSST = DATASETS.fetch('marirtsstM1.b1.20190320.000000.nc')
EXAMPLE_MFRSR = DATASETS.fetch('sgpmfrsr7nchE11.b1.20210329.070000.nc')
EXAMPLE_SURFSPECALB1MLAWER = DATASETS.fetch('nsasurfspecalb1mlawerC1.c1.20160609.080000.nc')
EXAMPLE_SIGMA_MPLV5 = DATASETS.fetch('201509021500.bi')
EXAMPLE_RL1 = DATASETS.fetch('sgprlC1.a0.20160131.000000.nc')
EXAMPLE_CO2FLX4M = DATASETS.fetch('sgpco2flx4mC1.b1.20201007.001500.nc')
EXAMPLE_SWATS = DATASETS.fetch('sgpswatsE8.b1.20071229.000700.cdf')
EXAMPLE_SIRS = DATASETS.fetch('sgpsirsE13.b1.20190101.000000.cdf')
EXAMPLE_SIRS_SIRI_QC = DATASETS.fetch('sgpsirsC1.b1.20040101.000000.cdf')
EXAMPLE_GML_RADIATION = DATASETS.fetch('brw21001.dat')
EXAMPLE_GML_MET = DATASETS.fetch('met_brw_insitu_1_obop_hour_2020.txt')
EXAMPLE_GML_OZONE = DATASETS.fetch('brw_12_2020_hour.dat')
EXAMPLE_GML_CO2 = DATASETS.fetch('co2_brw_surface-insitu_1_ccgg_MonthlyData.txt')
EXAMPLE_GML_HALO = DATASETS.fetch('brw_CCl4_Day.dat')
EXAMPLE_MET_TEST1 = DATASETS.fetch('sgpmet_no_time.nc')
EXAMPLE_MET_TEST2 = DATASETS.fetch('sgpmet_test_time.nc')
EXAMPLE_NOAA_PSL = DATASETS.fetch('ctd21125.15w')
EXAMPLE_NOAA_PSL_TEMPERATURE = DATASETS.fetch('ctd22187.00t.txt')
EXAMPLE_SP2B = DATASETS.fetch('mosaossp2M1.00.20191216.130601.raw.20191216x193.sp2b')
EXAMPLE_INI = DATASETS.fetch('mosaossp2M1.00.20191216.000601.raw.20191216000000.ini')
EXAMPLE_HK = DATASETS.fetch('mosaossp2auxM1.00.20191217.010801.raw.20191216000000.hk')
EXAMPLE_MET_YAML = DATASETS.fetch('sgpmetE13.b1.yaml')
EXAMPLE_CLOUDPHASE = DATASETS.fetch('nsacloudphaseC1.c1.20180601.000000.nc')
EXAMPLE_AAF_ICARTT = DATASETS.fetch('AAFNAV_COR_20181104_R0.ict')
EXAMPLE_NEON = DATASETS.fetch(
    'NEON.D18.BARR.DP1.00002.001.000.010.001.SAAT_1min.2022-10.expanded.20221107T205629Z.csv'
)
EXAMPLE_NEON_VARIABLE = DATASETS.fetch('NEON.D18.BARR.DP1.00002.001.variables.20221201T110553Z.csv')
EXAMPLE_NEON_POSITION = DATASETS.fetch(
    'NEON.D18.BARR.DP1.00002.001.sensor_positions.20221107T205629Z.csv'
)
EXAMPLE_DOD = DATASETS.fetch('vdis.b1')
EXAMPLE_EBBR1 = DATASETS.fetch('sgp30ebbrE32.b1.20191125.000000.nc')
EXAMPLE_EBBR2 = DATASETS.fetch('sgp30ebbrE32.b1.20191130.000000.nc')
EXAMPLE_EBBR3 = DATASETS.fetch('sgp30ebbrE13.b1.20190601.000000.nc')
EXAMPLE_ECOR = DATASETS.fetch('sgp30ecorE14.b1.20190601.000000.cdf')
EXAMPLE_SEBS = DATASETS.fetch('sgpsebsE14.b1.20190601.000000.cdf')
EXAMPLE_SEBS_E39 = DATASETS.fetch('sgpsebsE39.b1.20230601.000000.cdf')
EXAMPLE_ECORSF_E39 = DATASETS.fetch('sgpecorsfE39.b1.20230601.000000.nc')
EXAMPLE_STAMP_E39 = DATASETS.fetch('sgpstampE39.b1.20230601.000000.nc')
EXAMPLE_STAMPPCP_E39 = DATASETS.fetch('sgpstamppcpE39.b1.20230601.000000.nc')
EXAMPLE_AMC_E39 = DATASETS.fetch('sgpamcE39.b1.20230601.000000.nc')
EXAMPLE_MFAS_SODAR = DATASETS.fetch('sodar.20230404.mnd')
EXAMPLE_ENA_MET = DATASETS.fetch('enametC1.b1.20221109.000000.cdf')
EXAMPLE_CCN = DATASETS.fetch('sgpaosccn2colaE13.b1.20170903.000000.nc')
EXAMPLE_OLD_QC = DATASETS.fetch('sgp30ecorE6.b1.20040705.000000.cdf')
EXAMPLE_SONDE_WILDCARD = DATASETS.fetch('sgpsondewnpnC1.b1.20190101.053200.cdf')
EXAMPLE_CEIL_WILDCARD = DATASETS.fetch('sgpceilC1.b1.20190101.000000.nc')
EXAMPLE_HYSPLIT = DATASETS.fetch('houstonaug300.0summer2010080100')
EXAMPLE_AOSACSM = DATASETS.fetch('sgpaosacsmE13.b2.20230420.000109.nc')

# Multiple files in a list
dlppi_multi_list = ['sgpdlppiC1.b1.20191015.120023.cdf', 'sgpdlppiC1.b1.20191015.121506.cdf']
EXAMPLE_DLPPI_MULTI = [DATASETS.fetch(file) for file in dlppi_multi_list]
noaa_psl_list = ['ayp22199.21m', 'ayp22200.00m']
EXAMPLE_NOAA_PSL_SURFACEMET = [DATASETS.fetch(file) for file in noaa_psl_list]
met_wildcard_list = [
    'sgpmetE13.b1.20190101.000000.cdf',
    'sgpmetE13.b1.20190102.000000.cdf',
    'sgpmetE13.b1.20190103.000000.cdf',
    'sgpmetE13.b1.20190104.000000.cdf',
    'sgpmetE13.b1.20190105.000000.cdf',
    'sgpmetE13.b1.20190106.000000.cdf',
    'sgpmetE13.b1.20190107.000000.cdf',
]
EXAMPLE_MET_WILDCARD = [DATASETS.fetch(file) for file in met_wildcard_list]
met_contour_list = [
    'sgpmetE15.b1.20190508.000000.cdf',
    'sgpmetE31.b1.20190508.000000.cdf',
    'sgpmetE32.b1.20190508.000000.cdf',
    'sgpmetE33.b1.20190508.000000.cdf',
    'sgpmetE34.b1.20190508.000000.cdf',
    'sgpmetE35.b1.20190508.000000.cdf',
    'sgpmetE36.b1.20190508.000000.cdf',
    'sgpmetE37.b1.20190508.000000.cdf',
    'sgpmetE38.b1.20190508.000000.cdf',
    'sgpmetE39.b1.20190508.000000.cdf',
    'sgpmetE40.b1.20190508.000000.cdf',
    'sgpmetE9.b1.20190508.000000.cdf',
    'sgpmetE13.b1.20190508.000000.cdf',
]
EXAMPLE_MET_CONTOUR = [DATASETS.fetch(file) for file in met_contour_list]
twp_sonde_wildcard_list = [
    'twpsondewnpnC3.b1.20060119.050300.custom.cdf',
    'twpsondewnpnC3.b1.20060119.112000.custom.cdf',
    'twpsondewnpnC3.b1.20060119.163300.custom.cdf',
    'twpsondewnpnC3.b1.20060119.231600.custom.cdf',
    'twpsondewnpnC3.b1.20060120.043800.custom.cdf',
    'twpsondewnpnC3.b1.20060120.111900.custom.cdf',
    'twpsondewnpnC3.b1.20060120.170800.custom.cdf',
    'twpsondewnpnC3.b1.20060120.231500.custom.cdf',
    'twpsondewnpnC3.b1.20060121.051500.custom.cdf',
    'twpsondewnpnC3.b1.20060121.111600.custom.cdf',
    'twpsondewnpnC3.b1.20060121.171600.custom.cdf',
    'twpsondewnpnC3.b1.20060121.231600.custom.cdf',
    'twpsondewnpnC3.b1.20060122.052600.custom.cdf',
    'twpsondewnpnC3.b1.20060122.111500.custom.cdf',
    'twpsondewnpnC3.b1.20060122.171800.custom.cdf',
    'twpsondewnpnC3.b1.20060122.232600.custom.cdf',
    'twpsondewnpnC3.b1.20060123.052500.custom.cdf',
    'twpsondewnpnC3.b1.20060123.111700.custom.cdf',
    'twpsondewnpnC3.b1.20060123.171600.custom.cdf',
    'twpsondewnpnC3.b1.20060123.231500.custom.cdf',
    'twpsondewnpnC3.b1.20060124.051500.custom.cdf',
    'twpsondewnpnC3.b1.20060124.111800.custom.cdf',
    'twpsondewnpnC3.b1.20060124.171700.custom.cdf',
    'twpsondewnpnC3.b1.20060124.231500.custom.cdf',
]
EXAMPLE_TWP_SONDE_WILDCARD = [DATASETS.fetch(file) for file in twp_sonde_wildcard_list]
twp_sonde_20060121_list = [
    'twpsondewnpnC3.b1.20060121.051500.custom.cdf',
    'twpsondewnpnC3.b1.20060121.111600.custom.cdf',
    'twpsondewnpnC3.b1.20060121.171600.custom.cdf',
    'twpsondewnpnC3.b1.20060121.231600.custom.cdf',
]
EXAMPLE_TWP_SONDE_20060121 = [DATASETS.fetch(file) for file in twp_sonde_20060121_list]
stamp_wildcard_list = [
    'sgpstampE13.b1.20200101.000000.nc',
    'sgpstampE31.b1.20200101.000000.nc',
    'sgpstampE32.b1.20200101.000000.nc',
    'sgpstampE33.b1.20200101.000000.nc',
    'sgpstampE34.b1.20200101.000000.nc',
    'sgpstampE9.b1.20200101.000000.nc',
]
EXAMPLE_STAMP_WILDCARD = [DATASETS.fetch(file) for file in stamp_wildcard_list]
mmcr_list = ['sgpmmcrC1.b1.1.cdf', 'sgpmmcrC1.b1.2.cdf']
EXAMPLE_MMCR = [DATASETS.fetch(file) for file in mmcr_list]
EXAMPLE_GML_AEROSOL_NAS = DATASETS.fetch(
    'US1200R.20200101000000.20210214053818.nephelometer.aerosol_light_scattering_coefficient.pm10.1y.1h.US06L_TSI_3563_MLO.US06L_scat_coef.lev2.nas'
)



try:
    import pysp2

    PYSP2_AVAILABLE = True
except ImportError:
    PYSP2_AVAILABLE = False


def read_hk_file(file_name):
    """
    This procedure will read in an SP2 housekeeping file and then
    store the timeseries data into a pandas DataFrame.

    Parameters
    ----------
    file_name: str
        The file name to read in

    Returns
    -------
    hk_ds: xarray.Dataset
        The housekeeping information in an xarray Dataset
    """
    if PYSP2_AVAILABLE:
        return pysp2.io.read_hk_file(file_name)
    else:
        raise ModuleNotFoundError(
            'PySP2 must be installed in order to read SP2 data and housekeeping files.'
        )


def read_sp2(file_name, debug=False, arm_convention=True):
    """
    Loads a binary SP2 raw data file and returns all of the wave forms
    into an xarray Dataset.

    Parameters
    ----------
    file_name: str
        The name of the .sp2b file to read.
    debug: bool
        Set to true for verbose output.
    arm_convention: bool
        If True, then the file name will follow ARM standard naming conventions.
        If False, then the file name follows the SP2 default naming convention.

    Returns
    -------
    ds: xarray.Dataset
        The xarray Dataset containing the raw SP2 waveforms for each particle.
    """
    if PYSP2_AVAILABLE:
        return pysp2.io.read_sp2(file_name, debug, arm_convention)
    else:
        raise ModuleNotFoundError(
            'PySP2 must be installed in order to read SP2 data and housekeeping files.'
        )


def read_sp2_dat(file_name, type):
    """
    This reads the .dat files that generate the intermediate parameters used
    by the Igor processing. Wildcards are supported.

    Parameters
    ----------
    file_name: str
        The name of the file to save to. Use a wildcard to open multiple files at once.
    type: str
        This parameter must be one of:
            'particle': Load individual particle timeseries from .dat file
            'conc': Load timeseries of concentrations.
    Returns
    -------
    ds: xarray.Dataset
        The xarray dataset to store the parameters in.
    """
    if PYSP2_AVAILABLE:
        return pysp2.io.read_dat(file_name, type)
    else:
        raise ModuleNotFoundError(
            'PySP2 must be installed in order to read SP2 data and housekeeping files.'
        )



"""
Plot a timeseries of sounding data
----------------------------------------------------

This is a simple example for how to plot a timeseries of sounding
data from the ARM SGP site.

Author: Robert Jackson
"""

from arm_test_data import DATASETS
from matplotlib import pyplot as plt

import act

filename_sonde = DATASETS.fetch('sgpsondewnpnC1.b1.20190101.053200.cdf')
sonde_ds = act.io.arm.read_arm_netcdf(filename_sonde)
print(sonde_ds)

# Plot data
display = act.plotting.TimeSeriesDisplay(sonde_ds)
display.add_subplots((3,), figsize=(15, 10))
display.plot('wspd', subplot_index=(0,))
display.plot('tdry', subplot_index=(1,))
display.plot('rh', subplot_index=(2,))
plt.show()



import numpy as np

import act


def test_get_stability_indices():
    sonde_ds = act.io.arm.read_arm_netcdf(act.tests.sample_files.EXAMPLE_SONDE1)
    sonde_ds = act.retrievals.calculate_stability_indicies(
        sonde_ds, temp_name='tdry', td_name='dp', p_name='pres'
    )
    np.testing.assert_allclose(
        sonde_ds['parcel_temperature'].values[0:5],
        [269.85, 269.745276, 269.678006, 269.622444, 269.572331],
        rtol=1e-5,
    )
    assert sonde_ds['parcel_temperature'].attrs['units'] == 'kelvin'
    np.testing.assert_almost_equal(sonde_ds['surface_based_cape'], 0.96, decimal=2)
    assert sonde_ds['surface_based_cape'].attrs['units'] == 'J/kg'
    assert sonde_ds['surface_based_cape'].attrs['long_name'] == 'Surface-based CAPE'
    np.testing.assert_almost_equal(sonde_ds['surface_based_cin'], 0.000, decimal=3)
    assert sonde_ds['surface_based_cin'].attrs['units'] == 'J/kg'
    assert sonde_ds['surface_based_cin'].attrs['long_name'] == 'Surface-based CIN'
    np.testing.assert_almost_equal(sonde_ds['most_unstable_cape'], 0.000, decimal=3)
    assert sonde_ds['most_unstable_cape'].attrs['units'] == 'J/kg'
    assert sonde_ds['most_unstable_cape'].attrs['long_name'] == 'Most unstable CAPE'
    np.testing.assert_almost_equal(sonde_ds['most_unstable_cin'], 0.000, decimal=3)
    assert sonde_ds['most_unstable_cin'].attrs['units'] == 'J/kg'
    assert sonde_ds['most_unstable_cin'].attrs['long_name'] == 'Most unstable CIN'

    np.testing.assert_almost_equal(sonde_ds['lifted_index'], 28.4, decimal=1)
    assert sonde_ds['lifted_index'].attrs['units'] == 'kelvin'
    assert sonde_ds['lifted_index'].attrs['long_name'] == 'Lifted index'
    np.testing.assert_equal(sonde_ds['level_of_free_convection'], np.array(np.nan))
    assert sonde_ds['level_of_free_convection'].attrs['units'] == 'hectopascal'
    assert sonde_ds['level_of_free_convection'].attrs['long_name'] == 'Level of free convection'
    np.testing.assert_almost_equal(
        sonde_ds['lifted_condensation_level_temperature'], -8.07, decimal=2
    )
    assert sonde_ds['lifted_condensation_level_temperature'].attrs['units'] == 'degree_Celsius'
    assert (
        sonde_ds['lifted_condensation_level_temperature'].attrs['long_name']
        == 'Lifted condensation level temperature'
    )
    np.testing.assert_almost_equal(sonde_ds['lifted_condensation_level_pressure'], 927.1, decimal=1)
    assert sonde_ds['lifted_condensation_level_pressure'].attrs['units'] == 'hectopascal'
    assert (
        sonde_ds['lifted_condensation_level_pressure'].attrs['long_name']
        == 'Lifted condensation level pressure'
    )
    sonde_ds.close()


def test_calculate_precipitable_water():
    sonde_ds = act.io.arm.read_arm_netcdf(act.tests.sample_files.EXAMPLE_SONDE1)
    assert sonde_ds['tdry'].units == 'C', 'Temperature must be in Celsius'
    assert sonde_ds['rh'].units == '%', 'Relative Humidity must be a percentage'
    assert sonde_ds['pres'].units == 'hPa', 'Pressure must be in hPa'
    pwv_data = act.retrievals.calculate_precipitable_water(
        sonde_ds, temp_name='tdry', rh_name='rh', pres_name='pres'
    )
    np.testing.assert_almost_equal(pwv_data, 0.8028, decimal=3)
    sonde_ds.close()


def test_calculate_pbl_liu_liang():
    files = act.tests.sample_files.EXAMPLE_TWP_SONDE_20060121.copy()
    files2 = act.tests.sample_files.EXAMPLE_SONDE1
    files.append(files2)
    files.sort()

    pblht = []
    pbl_regime = []
    for file in files:
        ds = act.io.arm.read_arm_netcdf(file)
        ds['tdry'].attrs['units'] = 'degree_Celsius'
        ds = act.retrievals.sonde.calculate_pbl_liu_liang(ds, smooth_height=10)
        pblht.append(float(ds['pblht_liu_liang'].values))
        pbl_regime.append(ds['pblht_regime_liu_liang'].values)

    assert pbl_regime == ['NRL', 'NRL', 'NRL', 'NRL', 'NRL']
    np.testing.assert_array_almost_equal(pblht, [1038.4, 1079.0, 282.0, 314.0, 643.0], decimal=1)

    ds = act.io.arm.read_arm_netcdf(files[1])
    ds['tdry'].attrs['units'] = 'degree_Celsius'
    ds = act.retrievals.sonde.calculate_pbl_liu_liang(ds, land_parameter=False)
    np.testing.assert_almost_equal(ds['pblht_liu_liang'].values, 784, decimal=1)

    ds = act.io.arm.read_arm_netcdf(files[-2:])
    ds['tdry'].attrs['units'] = 'degree_Celsius'
    with np.testing.assert_raises(ValueError):
        ds = act.retrievals.sonde.calculate_pbl_liu_liang(ds)

    ds = act.io.arm.read_arm_netcdf(files[0])
    ds['tdry'].attrs['units'] = 'degree_Celsius'
    temp = ds['tdry'].values
    temp[10:20] = 19.0
    temp[0:10] = -10
    ds['tdry'].values = temp
    ds = act.retrievals.sonde.calculate_pbl_liu_liang(ds, land_parameter=False)
    assert ds['pblht_regime_liu_liang'].values == 'SBL'

    with np.testing.assert_raises(ValueError):
        ds2 = ds.where(ds['alt'].load() < 1000.0, drop=True)
        ds2 = act.retrievals.sonde.calculate_pbl_liu_liang(ds2, smooth_height=15)

    with np.testing.assert_raises(ValueError):
        ds2 = ds.where(ds['pres'].load() < 200.0, drop=True)
        ds2 = act.retrievals.sonde.calculate_pbl_liu_liang(ds2, smooth_height=15)

    with np.testing.assert_raises(ValueError):
        temp[0:5] = -40
        ds['tdry'].values = temp
        ds = act.retrievals.sonde.calculate_pbl_liu_liang(ds)

    ds = act.io.arm.read_arm_netcdf(files[0])
    ds['tdry'].attrs['units'] = 'degree_Celsius'
    temp = ds['tdry'].values
    temp[20:50] = 100.0
    ds['tdry'].values = temp
    with np.testing.assert_raises(ValueError):
        ds = act.retrievals.sonde.calculate_pbl_liu_liang(ds)


def test_calculate_heffter_pbl():
    files = act.tests.sample_files.EXAMPLE_TWP_SONDE_20060121.copy()
    files.sort()
    ds = act.io.arm.read_arm_netcdf(files[2])
    ds['tdry'].attrs['units'] = 'degree_Celsius'
    ds = act.retrievals.sonde.calculate_pbl_heffter(ds)
    assert ds['pblht_heffter'].values == 960.0
    np.testing.assert_almost_equal(ds['atm_pres_ss'].values[1], 994.9, 1)
    np.testing.assert_almost_equal(ds['potential_temperature_ss'].values[4], 298.4, 1)
    assert np.sum(ds['bottom_inversion'].values) == 7426.0
    assert np.sum(ds['top_inversion'].values) == 7903.0



"""
Spatial contour plot
--------------------

This is an example of how to prepare
and plot data for a contour plot

Author: Adam Theisen

"""


from arm_test_data import DATASETS
import matplotlib.pyplot as plt

import act

met_contour_list = [
    'sgpmetE15.b1.20190508.000000.cdf',
    'sgpmetE31.b1.20190508.000000.cdf',
    'sgpmetE32.b1.20190508.000000.cdf',
    'sgpmetE33.b1.20190508.000000.cdf',
    'sgpmetE34.b1.20190508.000000.cdf',
    'sgpmetE35.b1.20190508.000000.cdf',
    'sgpmetE36.b1.20190508.000000.cdf',
    'sgpmetE37.b1.20190508.000000.cdf',
    'sgpmetE38.b1.20190508.000000.cdf',
    'sgpmetE39.b1.20190508.000000.cdf',
    'sgpmetE40.b1.20190508.000000.cdf',
    'sgpmetE9.b1.20190508.000000.cdf',
    'sgpmetE13.b1.20190508.000000.cdf',
]

met_contour_filenames = [DATASETS.fetch(file) for file in met_contour_list]

time = '2019-05-08T04:00:00.000000000'
data = {}
fields = {}
wind_fields = {}
station_fields = {}
for f in met_contour_filenames:
    ds = act.io.arm.read_arm_netcdf(f)
    data.update({f: ds})
    fields.update({f: ['lon', 'lat', 'temp_mean']})
    wind_fields.update({f: ['lon', 'lat', 'wspd_vec_mean', 'wdir_vec_mean']})
    station_fields.update(
        {
            f: [
                'lon',
                'lat',
                'temp_mean',
                'atmos_pressure',
                'vapor_pressure_mean',
                'rh_mean',
            ]
        }
    )

display = act.plotting.ContourDisplay(data, figsize=(8, 8))
display.create_contour(fields=fields, time=time, levels=50)
display.plot_vectors_from_spd_dir(fields=wind_fields, time=time, mesh=True, grid_delta=(0.1, 0.1))
display.plot_station(fields=station_fields, time=time, markersize=7, color='red')
plt.show()



import numpy as np

from act.io.arm import read_arm_netcdf
from act.qc.arm import add_dqr_to_qc
from act.tests import EXAMPLE_ENA_MET, EXAMPLE_OLD_QC


def test_scalar_dqr():
    # Test DQR Webservice using known DQR
    ds = read_arm_netcdf(EXAMPLE_ENA_MET)

    # DQR webservice does go down, so ensure it
    # properly runs first before testing
    try:
        ds = add_dqr_to_qc(ds, assessment='Reprocessed,Suspect,Incorrect')
        ran = True
    except ValueError:
        ran = False

    if ran:
        assert 'qc_lat' in ds
        assert np.size(ds['qc_lon'].values) == 1
        assert np.size(ds['qc_lat'].values) == 1
        assert np.size(ds['qc_alt'].values) == 1
        assert np.size(ds['base_time'].values) == 1


def test_get_attr_info():
    ds = read_arm_netcdf(EXAMPLE_OLD_QC, cleanup_qc=True)
    assert 'flag_assessments' in ds['qc_lv'].attrs
    assert 'fail_min' in ds['qc_lv'].attrs
    assert ds['qc_lv'].attrs['flag_assessments'][0] == 'Bad'
    assert ds['qc_lv'].attrs['flag_masks'][-1] == 4



import numpy as np

import act


def test_calc_cog_sog():
    ds = act.io.arm.read_arm_netcdf(act.tests.sample_files.EXAMPLE_NAV)

    ds = act.utils.calc_cog_sog(ds)

    cog = ds['course_over_ground'].values
    sog = ds['speed_over_ground'].values

    np.testing.assert_almost_equal(cog[10], 170.987, decimal=3)
    np.testing.assert_almost_equal(sog[15], 0.448, decimal=3)

    ds = ds.rename({'lat': 'latitude', 'lon': 'longitude'})
    ds = act.utils.calc_cog_sog(ds)
    np.testing.assert_almost_equal(cog[10], 170.987, decimal=3)
    np.testing.assert_almost_equal(sog[15], 0.448, decimal=3)



from datetime import datetime
import pytest
import numpy as np

import act


def test_get_ord():
    time_window = [datetime(2020, 2, 4, 2, 0), datetime(2020, 2, 12, 10, 0)]
    my_asoses = act.discovery.get_asos_data(time_window, station='ORD', regions='IL')
    assert 'ORD' in my_asoses.keys()
    assert np.all(
        np.equal(
            my_asoses['ORD']['sknt'].values[:10],
            np.array([13.0, 11.0, 14.0, 14.0, 13.0, 11.0, 14.0, 13.0, 13.0, 13.0]),
        )
    )


def test_get_region():
    my_keys = ['MDW', 'IGQ', 'ORD', '06C', 'PWK', 'LOT', 'GYY']
    my_keys = sorted(my_keys)
    time_window = [datetime(2020, 2, 4, 2, 0), datetime(2020, 2, 12, 10, 0)]
    lat_window = (41.8781 - 0.5, 41.8781 + 0.5)
    lon_window = (-87.6298 - 0.5, -87.6298 + 0.5)
    with pytest.warns(UserWarning, match="No data available at station"):
        my_asoses = act.discovery.get_asos_data(
            time_window, lat_range=lat_window, lon_range=lon_window
        )
    asos_keys = sorted(list(my_asoses.keys()))
    assert asos_keys == my_keys



"""
Secondary Y-Axis Plotting
-------------------------

This example shows how to plot on the secondary y-axis
using Matplotlib functionality.  The secondary_y functionality has been removed from ACT.

"""

from arm_test_data import DATASETS
import matplotlib.pyplot as plt

import act

# Read in the data from a MET file
filename_met = DATASETS.fetch('sgpmetE13.b1.20190101.000000.cdf')
ds = act.io.arm.read_arm_netcdf(filename_met)

# Plot temperature and relative humidity with RH on the right axis
display = act.plotting.TimeSeriesDisplay(ds, figsize=(10, 6))

# Plot the data and make the y-axes color match the lines
display.plot('temp_mean', match_line_label_color=True)
display.day_night_background()

# Get the secondary y-axes and plot the RH on it
ax2 = display.axes[0].twinx()
ax2.plot(ds['time'], ds['rh_mean'], color='orange')

# Then the axes can be updated and modified through the normal matplotlib calls.
display.axes[0].set_yticks([-5, 0, 5])
display.axes[0].set_yticklabels(["That's cold", "Freezing", "Above Freezing"])

# Secondary y-axis will use the ax2 axes
ax2.set_yticks([65, 75, 85])
ax2.set_yticklabels(['Not as humid', 'Slightly Humid', 'Humid'])
ax2.set_ylabel('Relative Humidity (%)', color='orange')

plt.tight_layout()
plt.show()



import glob

import act


def test_read_neon():
    data_file = glob.glob(act.tests.EXAMPLE_NEON)
    variable_file = glob.glob(act.tests.EXAMPLE_NEON_VARIABLE)
    position_file = glob.glob(act.tests.EXAMPLE_NEON_POSITION)

    ds = act.io.neon.read_neon_csv(data_file)
    assert len(ds['time'].values) == 17280
    assert 'time' in ds
    assert 'tempSingleMean' in ds
    assert ds['tempSingleMean'].values[0] == -0.6003

    ds = act.io.neon.read_neon_csv(
        data_file, variable_files=variable_file, position_files=position_file
    )
    assert ds['northOffset'].values == -5.79
    assert ds['tempSingleMean'].attrs['units'] == 'celsius'
    assert 'lat' in ds
    assert ds['lat'].values == 71.282425



import act


def test_io_mpldataset():
    try:
        mpl_ds = act.io.mpl.read_sigma_mplv5(act.tests.EXAMPLE_SIGMA_MPLV5)
    except Exception:
        return

    # Tests fields
    assert 'channel_1' in mpl_ds.variables.keys()
    assert 'temp_0' in mpl_ds.variables.keys()
    assert mpl_ds.channel_1.values.shape == (102, 1000)

    # Tests coordinates
    assert 'time' in mpl_ds.coords.keys()
    assert 'range' in mpl_ds.coords.keys()
    assert mpl_ds.coords['time'].values.shape == (102,)
    assert mpl_ds.coords['range'].values.shape == (1000,)
    assert '_arm_standards_flag' in mpl_ds.attrs.keys()

    # Tests attributes
    assert '_datastream' in mpl_ds.attrs.keys()
    mpl_ds.close()



"""
Query the ARM DQR webservice
----------------------------

Simple example for querying the ARM DQR webservice
and plotting up the results

Author: Adam Theisen
"""

from arm_test_data import DATASETS
from matplotlib import pyplot as plt

import act

# Read in sample AOSMET data
filename_aosmet = DATASETS.fetch('maraosmetM1.a1.20180201.000000.nc')
ds = act.io.arm.read_arm_netcdf(filename_aosmet)

# Query DQR webservice for a specific variable
# As can be seen in the "Plotting QC Flags" example
# a call to obj.clean.cleanup() would normally be needed
# in order to plot up ARM's QC information.  In this case
# the call to add DQRs to the QC automatically applies that
# cleanup so you don't have to.
variable = 'temperature_ambient'
ds = act.qc.arm.add_dqr_to_qc(ds, variable=variable)

# Plot data
# Creat Plot Display
display = act.plotting.TimeSeriesDisplay(ds, figsize=(15, 10), subplot_shape=(2,))

# Plot temperature data in top plot
display.plot(variable, subplot_index=(0,))

# Plot QC data
display.qc_flag_block_plot(variable, subplot_index=(1,))
plt.show()



"""
Functions for solar radiation related calculations and retrievals.

"""

import numpy as np
import xarray as xr
from scipy.constants import Stefan_Boltzmann

from act.utils.geo_utils import get_solar_azimuth_elevation
from act.utils.data_utils import convert_units


def calculate_dsh_from_dsdh_sdn(
    ds,
    dsdh='down_short_diffuse_hemisp',
    sdn='short_direct_normal',
    lat='lat',
    lon='lon',
):
    """

    Function to derive the downwelling shortwave hemispheric irradiance (dsh) from the
    downwelling shortwave diffuse hemispheric irradiance (dsdh) and the shortwave
    direct normal irradiance (sdn). The derived values are added the returned
    Datasets as a new varible.

    Parameters
    ----------
    ds : xarray.Dataset
        Xarray dataset where variables for these calculations are stored.
    dsdh : str
        Name of the downwelling shortwave diffuse hemispheric irradiance field to use.
    sdn : str
        Name of shortwave direct normal irradiance field to use.
    lat : str
        Name of latitude variable in dataset to use for deriving solar zenight angle.
    lon : str
        Name of longitude variable in dataset to use for deriving solar zenight angle.

    Returns
    -------

    ds: xarray.Dataset
        ACT Xarray Dataset with calculations included as new variable.

    """

    ds = calculate_ghi_from_dni_dhi(ds, dni=sdn, dhi=dsdh, lat=lat, lon=lon)

    return ds


def calculate_ghi_from_dni_dhi(
    ds,
    dni='short_direct_normal',
    dhi='down_short_diffuse_hemisp',
    derived='derived_down_short_hemisp',
    long_name='Derived global horizontal irradiance',
    solar_zenith=None,
    lat='lat',
    lon='lon',
):
    """

    Function to derive the Global Horizontal Irradiance (GHI) from
    Direct Normal Irradiance (DNI) and Diffuse Horizontal
    Irradiance (DHI). The derived values are added the returned Datasets as a new DataArray.

    Parameters
    ----------
    ds : xarray.Dataset
        Xarray dataset where variables used for these calculations are stored
    dni : str
        Name of the direct normal irradiance DataArray to use.
    dhi : str
        Name of diffuse hemispheric irradiance DataArray to use.
    derived : str
        Name of new diffuse horizontal irradiance DataArray.
    long_name : str
        Long name used in new DataArray.
    solar_zenith : str or None
        Name of solar zenith DataArray in Dataset. If set to None will calculate using
        location and time variables from Dataset.
    lat : str
        Name of latitude DataArray in dataset to use for deriving solar zenight angle.
        Ignored if solar_zenith provided.
    lon : str
        Name of longitued DataArray in dataset to use for deriving solar zenight angle.
        Ignored if solar_zenith provided.

    Returns
    -------

    ds: xarray.Dataset
        ACT Xarray Dataset with global horizontal irradiance included as new DataArray.

    """

    # Get solar zenith angle
    if solar_zenith is not None:
        sz = ds[solar_zenith].values
        sz = convert_units(sz, ds[solar_zenith].attrs['units'], 'radians')
        cos_sz = np.cos(sz)
    else:
        elevation, _, _ = get_solar_azimuth_elevation(
            ds[lat].values, ds[lon].values, ds['time'].values
        )
        cos_sz = np.cos(np.radians(90.0 - elevation))

    ghi = ds[dhi].values + (cos_sz * ds[dni].values)

    # Add data into Dataset
    ds[derived] = xr.DataArray(
        ghi,
        dims=['time'],
        attrs={
            'long_name': long_name,
            'units': ds[dhi].attrs['units'],
        },
    )

    return ds


def calculate_dni_from_dhi_ghi(
    ds,
    dhi='down_short_diffuse_hemisp',
    ghi='down_short_hemisp',
    derived='derived_short_direct_normal',
    long_name='Derived direct normal irradiance',
    solar_zenith=None,
    lat='lat',
    lon='lon',
):
    """

    Function to derive the Direct Normal Irradiance (DNI) from
    Diffuse Horizontal Irradiance (DHI) and Global Horizontal
    Irradiance (GHI). The derived values are added the returned Datasets as a new DataArray.

    Parameters
    ----------
    ds : xarray.Dataset
        Xarray dataset where variables used for these calculations are stored
    dhi : str
        Name of the diffuse horizontal irradiance DataArray to use.
    ghi : str
        Name of global hemispheric irradiance DataArray to use.
    derived : str
        Name of new direct normal irradiance DataArray.
    long_name : str
        Long name used in new DataArray.
    solar_zenith : str or None
        Name of solar zenith DataArray in Dataset. If set to None will calculate using
        location and time variables from Dataset.
    lat : str
        Name of latitude DataArray in dataset to use for deriving solar zenight angle.
        Ignored if solar_zenith provided.
    lon : str
        Name of longitued DataArray in dataset to use for deriving solar zenight angle.
        Ignored if solar_zenith provided.

    Returns
    -------

    ds: xarray.Dataset
        ACT Xarray Dataset with direct normal irradiance included as new DataArray.

    """

    # Get solar zenith angle
    if solar_zenith is not None:
        sz = ds[solar_zenith].values
        sz = convert_units(sz, ds[solar_zenith].attrs['units'], 'radians')
        cos_sz = np.cos(sz)
    else:
        elevation, _, _ = get_solar_azimuth_elevation(
            ds[lat].values, ds[lon].values, ds['time'].values
        )
        cos_sz = np.cos(np.radians(90.0 - elevation))

    dni = (ds[ghi].values - ds[dhi].values) / cos_sz

    # Add data into Dataset
    ds[derived] = xr.DataArray(
        dni,
        dims=['time'],
        attrs={
            'long_name': long_name,
            'units': ds[dhi].attrs['units'],
        },
    )

    return ds


def calculate_dhi_from_dni_ghi(
    ds,
    dni='short_direct_normal',
    ghi='down_short_hemisp',
    derived='derived_down_short_diffuse_hemisp',
    long_name='Derived diffuse horizontal irradiance',
    solar_zenith=None,
    lat='lat',
    lon='lon',
):
    """

    Function to derive the Diffuse Horizontal Irradiance (DHI) from
    Direct Normal Irradiance (DNI) and Global Horizontal
    Irradiance (GHI). The derived values are added the returned Datasets as a new DataArray.

    Parameters
    ----------
    ds : xarray.Dataset
        Xarray dataset where variables used for these calculations are stored
    dni : str
        Name of the dirret normal irradiance DataArray to use.
    ghi : str
        Name of global hemispheric irradiance DataArray to use.
    derived : str
        Name of new diffuse horizontal irradiance DataArray.
    long_name : str
        Long name used in new DataArray.
    solar_zenith : str or None
        Name of solar zenith DataArray in Dataset. If set to None will calculate using
        location and time variables from Dataset.
    lat : str
        Name of latitude DataArray in dataset to use for deriving solar zenight angle.
        Ignored if solar_zenith provided.
    lon : str
        Name of longitued DataArray in dataset to use for deriving solar zenight angle.
        Ignored if solar_zenith provided.

    Returns
    -------

    ds: xarray.Dataset
        ACT Xarray Dataset with diffuse horizontal irradiance included as new DataArray.

    """

    # Get solar zenith angle
    if solar_zenith is not None:
        sz = ds[solar_zenith].values
        sz = convert_units(sz, ds[solar_zenith].attrs['units'], 'radians')
        cos_sz = np.cos(sz)
    else:
        elevation, _, _ = get_solar_azimuth_elevation(
            ds[lat].values, ds[lon].values, ds['time'].values
        )
        cos_sz = np.cos(np.radians(90.0 - elevation))

    dhi = ds[ghi].values - (ds[dni].values * cos_sz)

    # Add data into Dataset
    ds[derived] = xr.DataArray(
        dhi,
        dims=['time'],
        attrs={
            'long_name': long_name,
            'units': ds[ghi].attrs['units'],
        },
    )

    return ds


def calculate_irradiance_stats(
    ds,
    variable=None,
    variable2=None,
    diff_output_variable=None,
    ratio_output_variable=None,
    threshold=None,
):
    """

    Function to calculate the difference and ratio between two irradiance.

    Parameters
    ----------
    ds : xarray.Dataset
        Xarray dataset where variables for these calculations are stored
    variable : str
        Name of the first irradiance variable
    variable2 : str
        Name of the second irradiance variable
    diff_output_variable : str
        Variable name to store the difference results
        Defaults to 'diff_' + variable
    ratio_output_variable : str
        Variable name to store the ratio results
        Defaults to 'ratio_' + variable

    Returns
    -------

    ds : xarray.Dataset
        Xarray dataset with calculations included as new variables.

    """

    if variable is None or variable2 is None:
        return ds

    if diff_output_variable is None:
        diff_output_variable = 'diff_' + variable

    if ratio_output_variable is None:
        ratio_output_variable = 'ratio_' + variable

    # ---------------------------------
    # Calculating Difference
    # ---------------------------------
    diff = ds[variable] - ds[variable2]
    atts = {
        'long_name': ' '.join(['Difference between', variable, 'and', variable2]),
        'units': 'W/m^2',
    }
    da = xr.DataArray(diff, coords={'time': ds['time'].values}, dims=['time'], attrs=atts)
    ds[diff_output_variable] = da

    # ---------------------------------
    # Calculating Irradiance Ratio
    # ---------------------------------
    ratio = ds[variable].values / ds[variable2].values
    if threshold is not None:
        index = np.where((ds[variable].values < threshold) & (ds[variable2].values < threshold))
        ratio[index] = np.nan

    atts = {
        'long_name': ' '.join(['Ratio between', variable, 'and', variable2]),
        'units': '',
    }
    da = xr.DataArray(ratio, coords={'time': ds['time'].values}, dims=['time'], attrs=atts)
    ds[ratio_output_variable] = da

    return ds


def calculate_net_radiation(
    ds,
    ush='up_short_hemisp',
    ulh='up_long_hemisp',
    dsh='down_short_hemisp',
    dlhs='down_long_hemisp_shaded',
    smooth=None,
):
    """

    Function to calculate the net radiation from upwelling short and long-wave irradiance and
    downwelling short and long-wave hemisperic irradiances

    Parameters
    ----------
    ds : xarray.Dataset
        Xarray dataset where variables for these calculations are stored
    ush : str
        Name of the upwelling shortwave hemispheric variable
    ulh : str
        Name of the upwelling longwave hemispheric variable
    dsh : str
        Name of the downwelling shortwave hemispheric variable
    dlhs : str
        Name of the downwelling longwave hemispheric variable
    smooth : int
        Smoothing to apply to the net radiation.  This will create an additional variable

    Returns
    -------

    ds : xarray.Dataset
        Xarray dataset with calculations included as new variables.

    """

    # Calculate Net Radiation
    ush_da = ds[ush]
    ulh_da = ds[ulh]
    dsh_da = ds[dsh]
    dlhs_da = ds[dlhs]

    net = -ush_da + dsh_da - ulh_da + dlhs_da

    atts = {'long_name': 'Calculated Net Radiation', 'units': 'W/m^2'}
    da = xr.DataArray(net, coords={'time': ds['time'].values}, dims=['time'], attrs=atts)
    ds['net_radiation'] = da

    if smooth is not None:
        net_smoothed = net.rolling(time=smooth).mean()
        atts = {
            'long_name': 'Net Radiation Smoothed by ' + str(smooth),
            'units': 'W/m^2',
        }
        da = xr.DataArray(
            net_smoothed, coords={'time': ds['time'].values}, dims=['time'], attrs=atts
        )
        ds['net_radiation_smoothed'] = da

    return ds


def calculate_longwave_radiation(
    ds,
    temperature_var=None,
    vapor_pressure_var=None,
    met_ds=None,
    emiss_a=0.61,
    emiss_b=0.06,
):
    """

    Function to calculate longwave radiation during clear and cloudy sky conditions
    using equations from Monteith and Unsworth 2013, Prata 1996, as reported in
    Splitt and Bahrmann 1999.

    Parameters
    ----------
    ds : xarray.Dataset
        Xarray dataset where variables for these calculations are stored
    temperature_var : str
        Name of the temperature variable to use
    vapor_pressure_var : str
        Name of the vapor pressure variable to use
    met_ds : xarray.Dataset
        Xarray dataset where surface meteorological variables for these calculations are
        stored if not given, will assume they are in the main dataset passed in
    emiss_a : float
        a coefficient for the emissivity calculation of e = a + bT
    emiss_b : float
        a coefficient for the emissivity calculation of e = a + bT

    Returns
    -------
    ds : xarray.Dataset
        Xarray dataset with 3 new variables; monteith_clear, monteith_cloudy, prata_clear

    References
    ---------
    Monteith, John L., and Mike H. Unsworth. 2013. Principles of Environmental Physics.
        Edited by John L. Monteith and Mike H. Unsworth. Boston: Academic Press.

    Prata, A. J. 1996. “A New Long-Wave Formula for Estimating Downward Clear-Sky Radiation at
        the Surface.” Quarterly Journal of the Royal Meteorological Society 122 (533): 1127–51.

    Splitt, M. E., and C. P. Bahrmann. 1999. Improvement in the Assessment of SIRS Broadband
        Longwave Radiation Data Quality. Ninth ARM Science Team Meeting Proceedings,
        San Antonio, Texas, March 22-26

    """
    if met_ds is not None:
        T = met_ds[temperature_var] + 273.15  # C to K
        e = met_ds[vapor_pressure_var] * 10.0  # kpa to hpa
    else:
        T = ds[temperature_var] + 273.15  # C to K
        e = ds[vapor_pressure_var] * 10.0  # kpa to hpa

    if len(T) == 0 or len(e) == 0:
        raise ValueError('Temperature and Vapor Pressure are Needed')

    # Get Stefan Boltzmann Constant
    stefan = Stefan_Boltzmann

    # Calculate sky emissivity from Splitt and Bahrmann 1999
    esky = emiss_a + emiss_b * np.sqrt(e)

    # Base clear sky longwave calculation from Monteith 2013
    lw_calc_clear = esky * stefan * T**4

    # Prata 1996 Calculation
    xi = 46.5 * (e / T)
    lw_calc_clear_prata = (1.0 - (1.0 + xi) * np.exp(-((1.2 + 3.0 * xi) ** 0.5))) * stefan * T**4

    # Monteith Cloudy Calcuation as indicated by Splitt and Bahrmann 1999
    lw_calc_cldy = esky * (1.0 + (0.178 - 0.00957 * (T - 290.0))) * stefan * T**4

    atts = {'long_name': 'Clear Sky Estimate-(Monteith, 1973)', 'units': 'W/m^2'}
    da = xr.DataArray(lw_calc_clear, coords={'time': ds['time'].values}, dims=['time'], attrs=atts)
    ds['monteith_clear'] = da

    atts = {'long_name': 'Overcast Sky Estimate-(Monteith, 1973)', 'units': 'W/m^2'}
    da = xr.DataArray(lw_calc_cldy, coords={'time': ds['time'].values}, dims=['time'], attrs=atts)
    ds['monteith_cloudy'] = da

    atts = {'long_name': 'Clear Sky Estimate-(Prata, 1996)', 'units': 'W/m^2'}
    da = xr.DataArray(
        lw_calc_clear_prata,
        coords={'time': ds['time'].values},
        dims=['time'],
        attrs=atts,
    )
    ds['prata_clear'] = da

    return ds



"""
Forcing line plots with 2D data and QC
-----------------------------------------------------------

This is an example of how to use 2 dimentional DataArrays containing
multiple 1 dimentional data, including a summary quality control
plot showing the tests for each 1 dimentional data slice. This
example uses the force_line_plot keyword which will force multiple
1 dimentional plots for each element along the second dimention of
the non-time dimention.

Author: Ken Kehoe

"""

from arm_test_data import DATASETS
import matplotlib.pyplot as plt

from act.io.arm import read_arm_netcdf
from act.plotting import TimeSeriesDisplay

# Read a data file that has a 2D DataArray of multiple 1D data.
# The corresponding quality control DataArray is also read in and
# will be used to make a summary plot of quality control infomation
# of each assessment category.
filename_surf = DATASETS.fetch('nsasurfspecalb1mlawerC1.c1.20160609.080000.nc')
ds = read_arm_netcdf(filename_surf)

# The name of the data variable we wish to plot
var_name = 'surface_albedo_mfr_narrowband_10m'

# Create the ACT display object used for plotting. This will have two
# vertical plots of 800 by 400 pixels.
display = TimeSeriesDisplay(ds, subplot_shape=(2,), figsize=(8, 2 * 4))

# Create the top plot of data using the force_line_plot option.
# This will force the plotting to not assume the data are 2D data that
# would normally be plotted as a 2D plot. Rather for each index into the
# filter dimention plot a 1D time series plot. Setting labels=True
# will create a legend using the filter dimention DataArray.
display.plot(var_name, force_line_plot=True, labels=True)

# Create the bottom plot of summarized quality control by assessment
# cateory.
display.qc_flag_block_plot(var_name, subplot_index=(1,))

# Show the plot in a new window.
plt.show()



"""
Module that containing utilities involving datetimes.

"""

import datetime as dt
import warnings

import numpy as np
import pandas as pd
from scipy import stats


def dates_between(sdate, edate):
    """
    Ths procedure returns all of the dates between *sdate* and *edate*.

    Parameters
    ----------
    sdate : str
        The string containing the start date. The string is formatted
        YYYYMMDD.
    edate : str
        The string containing the end date. The string is formatted
        YYYYMMDD.

    Returns
    -------
    all_dates : array of datetimes
        The array containing the dates between *sdate* and *edate*.

    """
    days = dt.datetime.strptime(edate, '%Y%m%d') - dt.datetime.strptime(sdate, '%Y%m%d')
    all_dates = [
        dt.datetime.strptime(sdate, '%Y%m%d') + dt.timedelta(days=d) for d in range(days.days + 1)
    ]
    return all_dates


def numpy_to_arm_date(_date, returnTime=False):
    """
    Given a numpy datetime64, return an ARM standard date (yyyymmdd).

    Parameters
    ----------
    date : numpy.datetime64
        Numpy datetime64 date.
    returnTime : boolean
        If set to true, returns time instead of date

    Returns
    -------
    arm_date : string or None
        Returns an arm date.

    """
    from dateutil.parser._parser import ParserError
    from pandas._libs.tslibs.parsing import DateParseError

    try:
        date = pd.to_datetime(str(_date))
        if returnTime is False:
            date = date.strftime('%Y%m%d')
        else:
            date = date.strftime('%H%M%S')
    except (ParserError, DateParseError):
        date = None

    return date


def reduce_time_ranges(time, time_delta=60, broken_barh=False):
    """
    Given a time series, this function will return a list of tuples of time
    ranges representing the contineous times where no data is detected missing.

    Parameters
    ----------
    time : numpy datetime64 array
        The numpy array of date time values.
    time_delta : int
        The number of seconds to use as default time step in time array.
    broken_barh : boolean
        Option to return start time and duration instead of start time and
        end time. This is used with the pyplot.broken_barh() plotting routine.

    Returns
    -------
    time_ranges : list of tuples with 2 numpy datetime64 times
        The time range(s) of contineous data.

    """
    # Convert integer sections to numpy datetime64
    time_delta = np.timedelta64(int(time_delta * 1000), 'ms')

    # Make a difference array to find where time difference is great than time_delta
    diff = np.diff(time)
    dd = np.where(diff > time_delta)[0]

    if len(dd) == 0:
        return [(time[0], time[-1] - time[0])]

    # A add to start and end of array for beginning and end values
    dd = np.insert(dd, 0, -1)
    dd = np.append(dd, len(time) - 1)

    # Create a list of tuples containg time ranges or start time with duration
    if broken_barh:
        return [
            (time[dd[ii] + 1], time[dd[ii + 1]] - time[dd[ii] + 1]) for ii in range(len(dd) - 1)
        ]
    else:
        return [(time[dd[ii] + 1], time[dd[ii + 1]]) for ii in range(len(dd) - 1)]


def determine_time_delta(time, default=60):
    """
    Returns the most likely time step in seconds by analyzing the difference
    in time steps.

    Parameters
    ----------
    time : numpy datetime64 array
        The numpy array of date time values.
    default : int or float
        The default number to return if unable to calculate a value.

    Returns
    -------
    time_delta : float
        Returns the number of seconds for the most common time step. If can't
        calculate a value the default value is returned.

    """
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=RuntimeWarning)
        if time.size > 1:
            try:
                mode = stats.mode(np.diff(time), keepdims=True)
            except TypeError:
                mode = stats.mode(np.diff(time))
            time_delta = mode.mode[0]
            time_delta = time_delta.astype('timedelta64[s]').astype(float)
        else:
            time_delta = default

    return float(time_delta)


def datetime64_to_datetime(time):
    """
    Given a numpy datetime64 array time series, return datetime
    (y, m, d, h, m, s)

    Parameters
    ----------
    time : numpy datetime64 array, list of numpy datetime64 values or
        scalar numpy datetime64. The numpy array of date time values.

    Returns
    -------
    datetime : list
        Returns a list of datetimes (y, m, d, h, m, s) from a time series.
        YYYY-MM-DD, DD.MM.YYYY, DD/MM/YYYY or YYYYMMDD.

    """
    if isinstance(time, (tuple, list)):
        time = np.array(time)
    if len(time.shape) == 0:
        time = np.array([time])
    datetime_array = [
        dt.datetime.fromtimestamp(
            tm.astype('datetime64[ms]').astype('float') / 1000.0, tz=dt.timezone.utc
        ).replace(tzinfo=None)
        for tm in time
    ]
    return datetime_array


def date_parser(date_string, output_format='%Y%m%d', return_datetime=False):
    """Converts one datetime string to another or to
    a datetime object.

    Parameters
    ----------
    date_string : str
        datetime string to be parsed. Accepted formats are
        YYYY-MM-DD, DD.MM.YYYY, DD/MM/YYYY or YYYYMMDD.
    output_format : str
        Format for datetime.strftime to output datetime string.
    return_datetime : bool
        If true, returns str as a datetime object.
        Default is False.

    returns
    -------
    datetime_str : str
        A valid datetime string.
    datetime_obj : datetime.datetime
        A datetime object.

    """
    date_fmts = [
        '%Y-%m-%d',
        '%d.%m.%Y',
        '%d/%m/%Y',
        '%Y%m%d',
        '%Y/%m/%d',
        '%Y-%m-%dT%H:%M:%S',
        '%d.%m.%YT%H:%M:%S',
        '%d/%m/%YT%H:%M:%S',
        '%Y%m%dT%%H:%M:%S',
        '%Y/%m/%dT%H:%M:%S',
    ]
    for fmt in date_fmts:
        try:
            datetime_obj = dt.datetime.strptime(date_string, fmt)
            if return_datetime:
                return datetime_obj
            else:
                return datetime_obj.strftime(output_format)
        except ValueError:
            pass
    fmt_strings = ', '.join(date_fmts)
    raise ValueError('Invalid Date format, please use one of these formats ' + fmt_strings)


def adjust_timestamp(ds, time_bounds='time_bounds', align='left', offset=None):
    """
    Will adjust the timestamp based on the time_bounds or other information
    so that the timestamp aligns with user preference.

    Will work to adjust the times based on the time_bounds variable
    but if it's not available will rely on the user supplied input

    Parameters
    ----------
    ds : Xarray Dataset
        Dataset to adjust
    time_bounds : str
        Name of the time_bounds variable
    align : str
        Alignment of the time when using time_bounds.
            left: Sets timestamp to start of  sample interval
            right: Sets timestamp to end of sample interval
            center: Sets timestamp to middle of sample interval
    offset : int
        Time in seconds to offset the timestamp.  This overrides
        the time_bounds variable and can be positive or negative.
        Required to be in seconds

    Returns
    -------
    ds : Xarray DataSet
        Adjusted DataSet

    """

    if time_bounds in ds and offset is None:
        time_bounds = ds[time_bounds].values
        if align == 'left':
            time_start = [np.datetime64(t[0]) for t in time_bounds]
        elif align == 'right':
            time_start = [np.datetime64(t[1]) for t in time_bounds]
        elif align == 'center':
            time_start = [
                np.datetime64(t[0]) + (np.datetime64(t[0]) - np.datetime64(t[1])) / 2.0
                for t in time_bounds
            ]
        else:
            raise ValueError('Align should be set to one of [left, right, center]')

    elif offset is not None:
        time = ds['time'].values
        time_start = [t + np.timedelta64(offset, 's') for t in time]
    else:
        raise ValueError('time_bounds variable is not available')

    time_start = np.array(time_start).astype('datetime64[ns]')
    ds = ds.assign_coords({'time': time_start})

    return ds
